# =========================================================================
# Copyright (C) 2023. Huawei Technologies Co., Ltd. All rights reserved.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========================================================================


from fuxictr.tensorflow.tf_utils import get_initializer, get_regularizer
from tensorflow.keras.layers import Layer, Dense


class Linear(Layer):
    def __init__(self, 
                 output_dim, 
                 use_bias=True,
                 initializer="glorot_normal",
                 regularizer=None):
        super(Linear, self).__init__()
        self.linear = Dense(output_dim, use_bias=use_bias,
                            kernel_initializer=get_initializer(initializer),
                            kernel_regularizer=get_regularizer(regularizer),
                            bias_regularizer=get_regularizer(regularizer))
    
    def call(self, inputs):
        return self.linear(inputs)
